@@ -2,18 +2,13 @@
 #define CPUAFFINEDEFORMATIONFIELDKERNEL_H
 
 #include "AffineDeformationFieldKernel.h"
-#include "Content.h"
+#include "AladinContent.h"
 #include <string>
 
 
 class CPUAffineDeformationFieldKernel : public AffineDeformationFieldKernel {
 public:
-        CPUAffineDeformationFieldKernel(Content *con, std::string nameIn) : AffineDeformationFieldKernel(nameIn) {
-            this->deformationFieldImage = con->getCurrentDeformationField();
-            this->affineTransformation = con->getTransformationMatrix();
-            this->mask = con->getCurrentReferenceMask();
-        }
-
+        CPUAffineDeformationFieldKernel(AladinContent *con, std::string nameIn);
 
         void calculate(bool compose = false);
 
