@@ -1,10 +1,10 @@
 #include "CPUOptimiseKernel.h"
 
-CPUOptimiseKernel::CPUOptimiseKernel(Content *con, std::string name) : OptimiseKernel(name) {
+CPUOptimiseKernel::CPUOptimiseKernel(AladinContent *con, std::string name) : OptimiseKernel(name) {
     transformationMatrix = con->getTransformationMatrix();
     blockMatchingParams = con->getBlockMatchingParams();
 }
 
-void CPUOptimiseKernel::calculate(bool affine, bool ils, bool svd) {
+void CPUOptimiseKernel::calculate(bool affine) {
     optimize(this->blockMatchingParams, this->transformationMatrix, affine);
 }
